#include <stdlib.h>
#include <table.h>
#include <ndt.h>
#include <math.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>

typedef struct edge *edge;

struct edge {
  edge from_next;
  edge to_next;
  edge *from_last;
  edge *to_last;
  table t;
};

typedef struct node {
  char *name;
  edge from;
  edge to;
  table t;
} *node;

int each_property (eprop e, int *first)
{
  printf ("%c\"%s=%s\"",*first?',':' ',e->name,e->value);
  *first=1;
  return(1);
}

eprop allocate_prop(char *name,char *value)
{
  eprop e=(eprop)allocate_memory(sizeof(struct eprop));
  e->name=(char *)strdup(name);
  e->value=(char *)strdup(value);
  return(e);
}

void each_edge(table nodes,table t)
{
  eprop sparam=(eprop)table_find(t,"srcnode");
  eprop dparam=(eprop)table_find(t,"dstnode");
  eprop nodename=(eprop)table_find(t,"nodename");
  char *sname,*dname;
  node s,d;
  edge e;

  if(nodename){
    if (!(s=(node)table_find(nodes,nodename->value))){
      s=malloc(sizeof(struct node));
      s->name=strdup(nodename->value);
      s->from=0;
      s->to=0;
      table_insert(nodes,s,nodename->value);
    }
    s->t=t;
    return;
  }
  if (sparam && dparam){
    sname=sparam->value;
    dname=dparam->value;
    if (!(s=(node)table_find(nodes,sname))){
      s=malloc(sizeof(struct node));
      s->name=sname;
      s->from=0;
      s->to=0;
      table_insert(nodes,s,sname);
    }
    if (!(d=(node)table_find(nodes,dname))){
      d=malloc(sizeof(struct node));
      d->name=dname;
      d->from=0;
      d->to=0;
      table_insert(nodes,d,dname);
    }
    e=malloc(sizeof(struct edge));
    e->t=t;

    if (e->from_next=s->from) e->from_next->from_last=&e->from_next;
    *(e->from_last=&s->from)=e;

    if (e->to_next=d->to) e->to_next->to_last=&e->to_next;
    *(e->to_last=&d->to)=e;
  }
}

int node_compare(node n, char *name)
{
  return(strcmp(n->name,name)==0);
}

void dump_node(node n,void *k)
{
  edge i;
  int first;
  for (i=n->to;i;i=i->to_next){
    printf ("(");
    first=0;
    iterate_table_entries(i->t,each_property,&first);
    printf (")\n");
  }
}

int count_node(node n)
{
  int count=0,first=0;
  edge i;
  eprop z;
  char foo[10];

  for (i=n->from;i;i=i->from_next) count++;
  for (i=n->to;i;i=i->to_next) count++;
  sprintf (foo,"%d",count);
  table_insert(n->t,allocate_prop("degree",foo),"degree");
  printf ("(");
  iterate_table_entries(n->t,each_property,&first);
  printf (")\n");
}

main(int argc, char **argv)
{
  table nodes;
  void (*commit)()=0;
  void *commit_context;
  unsigned int i;
  int commit_count,commit_place;

  nodes=create_table(node_compare,key_from_string);
  for (i=1;i<argc;i++){ 
    if (argv[i][0]=='-'){
      if (commit){
	(*commit)(commit_count,argv[commit_place],argv[commit_place+1],
		  argv[commit_place+2],argv[commit_place+3]);
	commit=0;
      }
      commit_count=0;
      commit_place=i;
    } else commit_count++;
  }
  if (commit)
    (*commit)(commit_count,argv[commit_place],argv[commit_place+1],
	      argv[commit_place+2],argv[commit_place+3]);
  read_ndt_file(0,each_edge,nodes);
  iterate_table_entries(nodes, count_node,0 );
  iterate_table_entries(nodes, dump_node,0 );
}
