#include <table.h>
#include <ndt.h>

typedef struct edge *edge;
typedef struct node *node;

struct edge {
  edge from_next;
  edge to_next;
  edge *from_last;
  edge *to_last;
  node from,to;
  int live;
  table t;
};

struct node {
  char *name;
  edge from;
  edge to;
  int count;
  int number;
  table t;
};

int each_property (eprop e, int *first)
{
  printf ("%c\"%s=%s\"",*first?',':' ',e->name,e->value);
  *first=1;
  return(1);
}

eprop allocate_prop(char *name,char *value)
{
  eprop e=(eprop)allocate_memory(sizeof(struct eprop));
  e->name=(char *)strdup(name);
  e->value=(char *)strdup(value);
  return(e);
}

void insert_edge (node s,node d,table t)
{
  edge e=malloc(sizeof(struct edge));
  if (e->from_next=s->from) e->from_next->from_last=&e->from_next;
  *(e->from_last=&s->from)=e;
  
  if (e->to_next=d->to) e->to_next->to_last=&e->to_next;
  *(e->to_last=&d->to)=e;
  
  e->from=s;
  e->to=d;
  e->t=t;
  e->live=1;
}
static int eprop_compare(eprop p,char *n)
{
  return(string_equal(p->name,n));
}

node create_node(table t,char *name)
{
  node n;
  if (!(n=table_find(t,name))){
    n=malloc(sizeof(struct node));
    n->name=name;
    n->from=0;
    n->to=0;
    n->count=0;
    table_insert(t,n,name);
    n->t=create_table(eprop_compare,key_from_string);
  }
  return(n);
}

typedef struct reader {
  int state;
  char name[255];
  int pointer;
  node n;
  table t;
} *reader;

void parse_ged (struct reader *r,char *b, int length)
{
  int i;
  table t;
  node d;
  for (i=0;i<length;i++)
    switch (r->state){
    case 0:
      if (b[i]=='\n'){
	r->state=1;
	r->pointer=0;
      }
      break;
    case 1:
      if (b[i]=='E') {
	r->state=0;
	break;
      }
      if (b[i]==' '){
	r->name[r->pointer]=0;
	r->n=create_node(r->t,strdup(r->name));
	table_insert(r->n->t,allocate_prop("nodename",r->name),"nodename");
	r->pointer=0;
	r->state=2;
      } else if (b[i]!='\n') r->name[r->pointer++]=b[i];
      break;
    case 2:
      if (b[i]=='{') r->state=3;
      break;
    case 3:
      if (b[i]=='$') r->state=4;
      break;
    case 4:
      if (b[i]=='N') r->state=4;
    case 5:
      if (b[i]=='P') r->state=6;
      else r->state=4;
      break;
    case 6:
      if (b[i]==' ') r->state=7;
      else r->state=4;
      break;
    case 7:
      if (b[i]==' '){
	r->name[r->pointer]=0;
	table_insert(r->n->t,allocate_prop("x",r->name),"x");
	r->pointer=0;
	r->state=8;
      } else r->name[r->pointer++]=b[i];
      break;
    case 8:
      if (b[i]=='$'){
	r->name[r->pointer]=0;
	table_insert(r->n->t,allocate_prop("y",r->name),"y");
	r->pointer=0;
	r->state=9;
      } else r->name[r->pointer++]=b[i];
      break;
    case 9:
      if (b[i]=='}') r->state=10;
      break;
    case 10:
      if (b[i]==' ') r->state=11;
      break;
    case 11:
      if (b[i]=='"') r->state=12;
      break;
    case 12:
      if (b[i]=='"') {
	r->name[r->pointer]=0;
	/* node name? */
	r->pointer=0;
	r->state=13;
      } else r->name[r->pointer++]=b[i];
      break;
    case 13:
      r->pointer=0;
      if (b[i]==';'){
	r->state=1;
	break;
      }
      if (b[i]=='\n'){
	r->state=14;
	break;
      }
    case 14:
      if (b[i]==' '){
	r->name[r->pointer]=0;
	t=create_table(eprop_compare,key_from_string);
	d=create_node(r->t,strdup(r->name));
	insert_edge(r->n,d,t);
	table_insert(t,allocate_prop("dstnode",d->name),"dstnode");
	table_insert(t,allocate_prop("srcnode",r->n->name),"srcnode");
	/* edge number */
	r->pointer=0;
	r->state=15;
      } else r->name[r->pointer++]=b[i];
      break;
    case 15:
      if (b[i]=='"') {
	/* edge name */
	r->name[r->pointer]=0;
	r->pointer=0;
	r->state=13;
      } else r->name[r->pointer++]=b[i];
      break;
    }
}

int node_compare(node n, char *name)
{
  return(strcmp(n->name,name)==0);
}

void dump_edges(node n,void *k)
{
  edge i;
  int first;
  for (i=n->to;i;i=i->to_next){
    printf ("(");
    first=0;
    iterate_table_entries(i->t,each_property,&first);
    printf (")\n");
  }
}

void dump_node(node n,void *k)
{
  int first;
  printf ("(");
  first=0;
  iterate_table_entries(n->t,each_property,&first);
  printf (")\n");
}

main()
{
  struct reader r;
  char buffer[512];
  int i;
  r.t=create_table(node_compare,key_from_string);
  while ((i=read(0,buffer,sizeof(buffer)))>0)
    parse_ged(&r,buffer,i);
  iterate_table_entries(r.t,dump_node,1);
  iterate_table_entries(r.t,dump_edges,1);
}
