#include <table.h>
#include <ndt.h>
typedef struct parse_frame *parse_frame;

#define _inline_ inline

struct parse_frame { 
    parse_frame parent;
    void (*parent_function)(void *,void *,int,char);
    void *parent_context;
    void (*each_function)(parse_frame *,char);
    void *each_context;
    unsigned int index;
    unsigned int fill_pointer;
    unsigned int size;
    unsigned int count;
};

static _inline_ void char_insert(parse_frame p,char n)
{
  char *z;
  if (!(z=(char *)p->each_context))
    z=(char*)(p->each_context=(void *)allocate_memory(p->size=10));
  if (p->fill_pointer > (p->size-2))
    z=(char*)(p->each_context=
	      (void *)reallocate_memory(p->each_context,p->size+=10));
  z[p->fill_pointer++]=n;
}

void gather_argv(parse_frame *s,char n); 
void parse_string();

typedef struct edge_plist { 
  void (*finished)();
  void *context;
  table t;
} *edge_plist;

static int eprop_compare(eprop p,char *n)
{
  return(string_equal(p->name,n));
}


static _inline_ void frame_push(parse_frame *s,
		void (*callback)(void *,void *,int,char),
		void *context,
		void (*each)(parse_frame *,char),
		void *each_context)
{
    parse_frame p=(parse_frame)allocate_memory(sizeof(struct parse_frame));
    p->each_function=each;
    p->each_context=each_context;
    p->parent_function=callback;
    p->parent_context=context;
    p->fill_pointer=p->size=0;
    p->index=0;
    p->parent=*s;
    p->count=0;
    *s=p;
}


static _inline_ void frame_pop(parse_frame *s, char last)
{
    parse_frame p=*s;
    *s=(*s)->parent;
    if (p->parent_function) (*p->parent_function)(p->parent_context,
						  p->each_context,
						  p->fill_pointer,last);
    deallocate_memory(p);
}



static void eprop_name_finish(eprop p,char *name,int fill,char n)
{
  name[fill]=0;
  p->name=name;
}

static void eprop_value_finish(eprop p,char *value,int fill,char n)
{
  if (value) value[fill]=0;
  p->value=value;
  table_insert(p->t,p,p->name);
}

static void gather_word(parse_frame *s,char n)
{
  eprop p;
  switch (n){
  case '=':
    p=(eprop)(*s)->parent_context;
    frame_pop(s,n);
    frame_push(s,eprop_value_finish,p,gather_word,0);
    break;
  case '"':
    frame_pop(s,n);
    break;
  default:
    char_insert(*s,n);
  }
} 

static int free_edge(eprop p,void *foo)
{
  deallocate_memory(p->name);
  deallocate_memory(p->value);
  deallocate_memory(p);
  return(1);
}

void destroy_ndt_edge(table t)
{
    iterate_table_entries(t,free_edge,0);
    destroy_table(t);
}

static void ndt_create_edge(edge_plist e,eprop last,int fill, char n)
{
  (*e->finished)(e->context,e->t);
}

static void gather_properties(parse_frame *s, char n)
{
  eprop p;
  edge_plist e;
  switch (n){
  case ',':
    break;
  case '"':
    e=(edge_plist)(*s)->each_context;
    p=(eprop)allocate_memory(sizeof(struct eprop));
    p->t=e->t;
    frame_push(s,eprop_name_finish,p,gather_word,0);
    break;
  case ')':
    frame_pop(s,n);
  case ' ':
  case '\n':
    return;
  default:
    printf ("ndt parse error\n");
  }
}

void edge_delimit(parse_frame *s, char r)
{
  table t;
  edge_plist e,e0;
  switch(r){
  case '\n':
  case '\t':
  case ' ':
    break;
  default:
    if (r!='(') {
      printf ("ndt syntax\n");
    } else {
      e0=(edge_plist)(*s)->each_context;
      e=(edge_plist)allocate_memory(sizeof(struct edge_plist));
      e->context=e0->context;
      e->finished=e0->finished;
      e->t=create_table(eprop_compare,key_from_string);
      frame_push(s, ndt_create_edge, e, gather_properties,(void *) e);
    }
  }
}

void read_ndt_file(int fd,void (*each_edge)(),void *context)
{
  char buf[512];
  struct edge_plist n;
  int readlen,i;
  parse_frame *pf=(parse_frame *)allocate_memory(sizeof(parse_frame));
  
  frame_push(pf,0,0,edge_delimit,&n);
  n.finished=each_edge;
  n.context=context;
  while ((readlen=read(fd,buf,512))>0)
    for (i=0;i<readlen;i++)
      (*(*pf)->each_function)(pf,buf[i]);
}
