#!/usr/local/bin/perl

########################################################################
# ndt2gv.perl:
# convert canonical edge/node representation to Geomview output
#
# input: lines of name/value pairs.
#   required: edges should have "srcnode" and "dstnode".
#             nodes should have "nodename" and "position".
#   optional: if any edge is colored all edges must be colored.
#             this is not true for nodes, can have some but not others
#
# output: Geomview Command Language or OOGL
#  
# tamara munzner
# 6/96
########################################################################

$drawstyle = "gem3d" unless $drawstyle = $ARGV[0];

# "gv" or "oogl" or "fast"
# oogl can be turned into vrml easily, 
# while gv can be highlighted with mbone.perl
# fast is just oogl with no hyperlink stuff, for efficient drawing

$outstyle = "gv" unless $outstyle = $ARGV[1];
$sphererad = "1.0" unless $sphererad = $ARGV[2];

$htmldoc = "$drawstyle.html" unless $htmldoc = $ARGV[5];

open(JUNK, ">junk");
open(ANCHORS, ">$drawstyle.html");

open(TOGV, ">-");
if ($outstyle eq "gv") {
    print TOGV "(progn \n";
} else {
    print TOGV "{LIST \n";
}

$nodename = 0;
undef($bigstr);

while (<STDIN>) {
    undef($color, $position, $nodename, $ap);
    undef($srcstr, $dststr, $linkname, $handlename, $sphere);

    # deal with a series of items that look like this:
    # ( "name=val", "name2=val2", ...)
    chop;
    if (/nodename/) {
	# it's a node
	@line = eval($_);
	foreach $item (@line) {
	    ($lhs,$rhs) = split('=', $item);
	    if (defined($lhs) && defined($rhs)) {
		eval "\$$lhs = \"$rhs\";";
	    }
	}
	if (!defined($position)) {
	    print JUNK "need position for $nodename\n";
	    next;
	} else {
	    $nodepos{$nodename} = $position;
	    $nodenum{$nodename} = $nodenum;
	    $handlename = sprintf("N%d", $nodename);
	    $linkname = "linkN$nodenum,$nodename";
	    $srcstr = "node $nodename $position";
	    $dststr = "";
	    $ap = "appearance {material {diffuse $color}}" if $color;
	    $sphere = "{ $ap SPHERE $sphererad $position}";
	    $bigstr .= ($outstyle eq "fast") ? $sphere :
		&gvlink($sphere, $handlename, $srcstr, $dststr, $linkname);
	    $nodenum++;
	}
    } else {
	# it's an edge. defer processing.
	$edges[$i++] = $_;
    }
}

&gvout("nodes", $bigstr);
undef($bigstr);

foreach $edge (@edges) {
    undef($color,$srcnode,$dstnode, $edgecolor, $vect);
    undef($srcstr, $dststr, $linkname, $handlename, $str);

    @line = eval($edge);
    foreach $item (@line) {
	($lhs,$rhs) = split('=', $item);
	if (defined($lhs) && defined($rhs)) {
	    eval "\$$lhs = \"$rhs\";";
	}
    }
    if (!defined($srcnode) || !defined($dstnode)) {
	print JUNK "need srcnode or dst node for this edge: $edge\n";
	next;
    }

    $edgecolor = $color if defined($color);

    if ($outstyle eq "fast") {
	$vectcolors .= "$edgecolor 1 \n" if ($edgecolor);
	$vectinds .= "$nodenum{$srcnode} $nodenum{$dstnode} \n";
	$vectpts .= "$nodepos{$srcnode}   $nodepos{$dstnode} \n";
	$edgenum++;
	next;
    }

    $vect = sprintf("VECT 1 2 %d 2 %d $nodepos{$srcnode}  $nodepos{$dstnode}  %s ", defined($edgecolor), defined($edgecolor), $edgecolor);
    # VECT object wants RGBA colors, we're given an RGB color
    $vect .= " 1" if defined($edgecolor); 

    $linkname = "link$count,$srcnode,$dstnode";
    $handlename = join("/",$srcnode,$dstnode);
    $srcstr = "src $srcnode $nodepos{$srcnode}";
    $dststr = "dst $dstnode $nodepos{$dstnode}";
    $str = &gvlink($vect, $handlename, $srcstr, $dststr, $linkname);
    $bigstr .= $str;

    $edgenum++;
}

if ($outstyle eq "fast") {
    $bigstr = sprintf("{VECT %d %d %d \n", $edgenum, 2*$edgenum,
		      defined($vectcolors) ? $edgenum : 0);
    for ($i = 0; $i < $edgenum; $i++) {
	$bigstr .= "2 ";
    }
    $bigstr .= "\n";
    for ($i = 0; $i < $edgenum; $i++) {
	$bigstr .= defined($vectcolors) ? " 1" : " 0";
    }
    $bigstr .= "\n";
    $bigstr .= "$vectpts\n$vectcolors\n}\n";
}

&gvout("edges", $bigstr);

if ($outstyle eq "gv") {
    print TOGV ") \n";
} else {
    print TOGV "} \n";
}

close TOGV;
close ANCHORS;
close JUNK;

########################################################################
#
# subroutines
#
########################################################################

sub gvlink {
    local($gstr, $handlename, $srcstr, $dststr, $linkname) = @_;
    undef(local($handle, $str));

    print ANCHORS "<hr><a name=$linkname> $srcstr<p>$dststr<p>";

    if ($outstyle eq "oogl") {
	$handle = "{ $gstr }"; 
    } elsif ($outstyle eq "gv") {
	#use handles, for highlighting selection with mbone.perl
	print TOGV "(hdefine geometry $handlename $gstr)\n";
	$handle = ": " . $handlename;
    }
    $str = " {LIST \
	     { COMMENT SRC MBONE {\"$srcstr\"}} \
	     { COMMENT DST MBONE {\"$dststr\"}} \
	     { COMMENT mbone HTITLE { $srcstr $dststr }} \
	     { COMMENT mbone HREF {\"$htmldoc#$linkname\"}} \
	     $handle \
	     } ";
}


sub gvout {
    local($name, $body) = @_;

    if ($outstyle eq "gv") {
	print TOGV "(geometry $name {LIST $body })\n";
    } elsif ($outstyle eq "oogl") {
	print TOGV "{LIST $body}\n";
    } else { # $outstyle eq "fast"
	print TOGV "$body\n";
    }
}
