typedef struct ip_route *ip_route;

struct ip_route {
  unsigned int destination;
  unsigned int destination_mask;
  void *data;
  int use_count;
  unsigned int bit_position;
  ip_route on,off;
  ip_route parent;
};

void *resolve_ip_route(ip_route o,unsigned int destination)
{
  ip_route here=o;
  int bpos;
  ip_route best=0;
  unsigned int length=0;
  
  do {
    bpos=here->bit_position;
    if ((here->destination_mask > length)&&
	(destination & here->destination_mask) == here->destination){
      best=here;
      length=here->destination_mask;
    }
    if (destination & (1<<here->bit_position)) 
      here=here->on;
    else 
      here=here->off;
  } while (bpos > here->bit_position);
  if (best) return(best->data);
  /* not sure we need this with the stack
     there is a problem resolving default if the high
     bit is set. this is unpleasant but deals with the
     issue assuming that there are no route conflicts with net 0 */
  if (destination&0x80000000) return(resolve_ip_route(o,0));
  return(0);
}

static void add_route_into_tree(ip_route *p,ip_route r)
{
  int i;
  ip_route *next;
  
  for (i=31;i&&(i>(*p)->bit_position);i--)
    if ((((*p)->destination) & (1<<i)) != (r->destination & (1<<i))){
      r->on=(r->destination & (1<<i))?r:*p;
      r->off=(r->destination & (1<<i))?*p:r;
      r->bit_position=i;
      *p=r;
      return;
    }
  next=(r->destination & (1<<(*p)->bit_position))?&((*p)->on):
    &((*p)->off);
  if ((*next) == (*p)){
    if (r->destination & (1<<((*p)->bit_position)))
      (*p)->on=r; else (*p)->off=r;
    for (;i;i--)
      if ((((*p)->destination) & (1<<i)) != (r->destination & (1<<i))){
	r->on=(r->destination & (1<<i))?r:*p;
	r->off=(r->destination & (1<<i))?*p:r;
	r->bit_position=i;
      }
  } else {
    add_route_into_tree(next,r);
  }
  /* identical insertion*/
}

void add_route(ip_route *n,unsigned int to,
	       unsigned int tomask,
	       void *data)

{
  unsigned int k=((1<<tomask)-1)<<(32-tomask);
  ip_route r=(ip_route)malloc(sizeof(struct ip_route));
  r->bit_position=31;
  if (tomask==32) k=-1;
  r->destination=to&k;
  r->destination_mask=k;
  r->use_count=0;
  r->on=r->off=r;
  r->parent=0;
  r->data=data;
  if (!*n){
    r->bit_position=0;
    r->on=r;
    r->off=r;
    *n=r;
  } else add_route_into_tree(n,r);
}

static void print_routing_table_internal(ip_route r)
{ 
  int i;
  char a[40], b[40], c[40];
  int al;
  sprintf(a,"%d.%d.%d.%d",
	  r->destination>>24,(r->destination>>16)&255,
	  (r->destination>>8)&255,r->destination&255);
  al=strlen(a);
  sprintf(b,"%d.%d.%d.%d",
	  r->destination_mask>>24,(r->destination_mask>>16)&255,
	  (r->destination_mask>>8)&255,r->destination_mask&255);
  printf ("%s",a);
  for (i=0;i<(16-al);i++) printf (" ");
  printf ("%s",b);
  printf ("\n");
  
  if (r->on->bit_position < r->bit_position)
    print_routing_table_internal(r->on);
  if (r->off->bit_position < r->bit_position)
    print_routing_table_internal(r->off);

}

void print_routing_table(ip_route t)
{
  printf ("destination     netmask\n");
  if (t)  print_routing_table_internal(t);
}
